(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Equational_Reasoning_Impl
  imports
    Equational_Reasoning
    Completion
    TRS.Trs_Impl
begin

definition "showsl_eqs \<equiv> showsl_trs' showsl showsl (STR ''equational system:'') (STR '' = '')"
definition "showsl_eq \<equiv> showsl_rule' showsl showsl (STR '' = '')"
definition showsl_ln :: "nat \<Rightarrow> showsl" where
  "showsl_ln i \<equiv> showsl_nl \<circ> showsl i \<circ> showsl_lit (STR '': '')"

function
  eq_proof_lines :: "('f:: showl, 'v:: showl) eq_proof \<Rightarrow> nat \<Rightarrow> (showsl \<times> nat \<times> ('f, 'v) rule)"
and eq_proofs_lines :: "('f:: showl, 'v:: showl) eq_proof list \<Rightarrow> nat \<Rightarrow> (showsl \<times> nat list \<times> ('f, 'v) term list \<times> ('f, 'v) term list)"
where
  "eq_proof_lines (Refl s) i = (showsl_ln (Suc i) \<circ> showsl_eq (s, s) \<circ> showsl_lit (STR '' [refl]''), Suc i, s, s)"
| "eq_proof_lines (Sym p) i = (
    let (s, i', l, r) = eq_proof_lines p i in
    (s \<circ> showsl_ln (Suc i') \<circ> showsl_eq (r, l) \<circ> showsl_lit (STR '' [sym '') \<circ> showsl i' \<circ> showsl_lit (STR '']''), Suc i', r, l))"
| "eq_proof_lines (Trans p1 p2) i = (
    let (s1, i1, s, t) = eq_proof_lines p1 i in
    let (s2, i2, u, v) = eq_proof_lines p2 i1 in
    (s1 \<circ> s2 \<circ>
      showsl_ln (Suc i2) \<circ> showsl_eq (s, v) \<circ> showsl_lit (STR '' [trans '') \<circ> showsl i1 \<circ> showsl_lit (STR '', '') \<circ> showsl i2 \<circ> showsl_lit (STR '']''), Suc i2, s, v))"
| "eq_proof_lines (Assm (l, r) \<sigma>) i = (
    let eq = (l \<cdot> \<sigma>, r \<cdot> \<sigma>) in
    (showsl_ln (Suc i) \<circ> showsl_eq eq \<circ> showsl_lit (STR '' [assm '') \<circ> showsl_eq (l, r) \<circ> showsl_lit (STR '']''), Suc i, eq))"
| "eq_proof_lines (Cong f ps) i = (
    let (s, is, ls, rs) = eq_proofs_lines ps i in
    let eq = (Fun f ls, Fun f rs) in
    let i' = last is in
    let is = butlast is in
    (s \<circ> showsl_ln (Suc i') \<circ> showsl_eq eq \<circ> showsl_list_gen showsl (STR '' [cong]'') (STR '' [cong '') (STR '', '') (STR '']'') is, Suc i', eq))"
| "eq_proofs_lines [] i = (id, [i], [], [])"
| "eq_proofs_lines (p#ps) i = (
    let (s1, i', l, r) = eq_proof_lines p i in
    let (s2, is, ls, rs) = eq_proofs_lines ps i' in
    (s1 \<circ> s2, i'#is, l#ls, r#rs))"
  by (pat_completeness) auto
termination by (lexicographic_order)

abbreviation showsl_eq_proof where "showsl_eq_proof p \<equiv> fst (eq_proof_lines p 0)"

definition
  check_estep ::
    "('f:: showl, 'v:: showl) equation list \<Rightarrow>
    pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
    showsl check"
where
  "check_estep E p rule l_to_r s t \<equiv> do {
    check (\<exists>r\<in>set E. rule =\<^sub>v r)
      (showsl_rule rule \<circ> showsl_lit (STR '' is not an equation of\<newline>'') \<circ> showsl_eqs E \<circ> showsl_nl);
    check (p \<in> poss s) (showsl_pos p \<circ> showsl_lit (STR '' is not a position of '') \<circ> showsl s \<circ> showsl_nl);
    check (p \<in> poss t) (showsl_pos p \<circ> showsl_lit (STR '' is not a position of '') \<circ> showsl t \<circ> showsl_nl);
    let C = ctxt_of_pos_term p s;
    let D = ctxt_of_pos_term p t;
    let u = subt_at s p;
    let v = subt_at t p;
    let rrule = (if l_to_r then rule else (snd rule, fst rule));
    let err = (showsl_lit (STR ''the term '') \<circ> showsl t
      \<circ> showsl_lit (STR '' does not result from a proper application of term '') \<circ> showsl s \<circ> showsl_lit (STR '' using equation\<newline>'')
      \<circ> showsl_eq rrule \<circ> showsl_lit (STR '' at position '') \<circ> showsl_pos p \<circ> showsl_nl);
    (case match_list Var [(fst rrule, u), (snd rrule, v)] of
      Some \<sigma> \<Rightarrow> check (C = D) err
    | None \<Rightarrow> error err)
  }"

lemma check_estep_sound:
  assumes ok: "isOK (check_estep E p r l_to_r s t)"
  shows "(s, t) \<in> (rstep (set E))\<^sup>\<leftrightarrow>"
proof -
  let ?C = "ctxt_of_pos_term p s"
  let ?D = "ctxt_of_pos_term p t"
  let ?r = "if l_to_r then r else (snd r,fst r)"
  let ?mlist = "[(subt_at s p, fst ?r), (subt_at t p, snd ?r)]"
  note ok = ok[unfolded check_estep_def Let_def, simplified]
  from ok
  have ps: "p \<in> poss s" and pt: "p \<in> poss t" and "\<exists>rule \<in> set E. r =\<^sub>v rule" by auto    
  then obtain rule where rule: "rule \<in> set E" and inst: "(fst r, snd r) =\<^sub>v (fst rule, snd rule)" by auto
  from eq_rule_mod_varsE[OF inst]
  obtain \<delta> where fst: "fst r = fst rule \<cdot> \<delta>" and snd: "snd r = snd rule \<cdot> \<delta>" by auto
  let ?E = "(set E)\<^sup>\<leftrightarrow>"
  have "\<exists> rule \<in> ?E. fst ?r = fst rule \<cdot> \<delta> \<and> snd ?r = snd rule \<cdot> \<delta>"
  proof (cases l_to_r)
    case True
    then show ?thesis using rule fst snd by auto
  next
    case False
    show ?thesis
      by (rule bexI[of _ "(snd rule,fst rule)"], insert rule fst snd False, auto)
  qed
  then obtain rule where rule: "rule \<in> ?E" and fst: "fst ?r = fst rule \<cdot> \<delta>"
    and snd: "snd ?r = snd rule \<cdot> \<delta>" by auto
  from ok
    obtain \<sigma> where some: "match_list Var [(fst ?r, subt_at s p), (snd ?r, subt_at t p)] = Some \<sigma>"
    by force
  from match_list_sound [OF this]
    have "fst ?r \<cdot> \<sigma> = s |_ p" and "snd ?r \<cdot> \<sigma> = t |_ p" by auto
  then have s: "s |_ p = fst rule \<cdot> \<delta> \<circ>\<^sub>s \<sigma>" and t: "t = ?D\<langle>snd rule \<cdot> \<delta> \<circ>\<^sub>s \<sigma>\<rangle>"
    using ctxt_supt_id[OF pt] fst snd by simp+
  from ok and some
    have DC: "?D = ?C" by simp
  have "(s,t) \<in> rstep_r_p_s ?E rule p (\<delta> \<circ>\<^sub>s \<sigma>)"
    unfolding rstep_r_p_s_def Let_def
    by (rule, unfold split t[unfolded DC], intro conjI, rule ps, rule rule,
    insert ctxt_supt_id[OF ps, unfolded s], auto)
  then have "(s,t) \<in> rstep ?E" unfolding rstep_iff_rstep_r_p_s 
    using surjective_pairing by metis
  then show ?thesis unfolding estep_sym_closure_conv .
qed

type_synonym  ('f, 'v) eseq = "(pos \<times> ('f, 'v) equation \<times> bool \<times> ('f, 'v) term) list"

fun
  check_conversion ::
    "('f:: showl, 'v:: showl) equation list \<Rightarrow> 
     ('f, 'v) eseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> showsl check"
where
  "check_conversion E [] s u = check (s = u) (
    showsl_lit (STR ''the last term of the conversion\<newline>'') \<circ> showsl s \<circ>
    showsl_lit (STR ''\<newline>does not correspond to the goal term\<newline>'') \<circ> showsl u \<circ> showsl_nl)"
| "check_conversion E ((p, r, l_to_r, t) # c) s u = do {
    check_estep E p r l_to_r s t;
    check_conversion E c t u
  }"

lemma check_conversion_sound:
  assumes ok: "isOK (check_conversion E eseq s t)"
  shows "(s, t) \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*"
  using ok
proof (induct eseq arbitrary: s t)
  case Nil
  then show ?case by auto
next
  case (Cons st eseq s u)
  obtain p r l_to_r t where st: "st = (p,r,l_to_r,t)" by (cases st, force+)
  from Cons(2)[unfolded st, simplified]
  have one: "isOK(check_estep E p r l_to_r s t)" and rec: "isOK(check_conversion E eseq t u)" by auto
  from check_estep_sound[OF one] Cons(1)[OF rec] show ?case unfolding conversion_def 
    by (rule converse_rtrancl_into_rtrancl)
qed

type_synonym ('f, 'v) subsumption_proof = "(('f, 'v) equation \<times> ('f, 'v) eseq) list"

fun
  check_subsumptions_guided :: "('f :: showl, 'v :: showl) equation list \<Rightarrow> ('f, 'v) subsumption_proof \<Rightarrow> showsl check"
where "check_subsumptions_guided E [] = succeed"
    | "check_subsumptions_guided E ((e, seq) # convs) = do {
         check_conversion E seq (fst e) (snd e)
             <+? (\<lambda> s. showsl_lit (STR ''problem in conversion for equation '') \<circ> showsl_eq e \<circ> showsl_nl \<circ> s);
         check_subsumptions_guided (e # E) convs
       }"

lemma check_subsumptions_guided:
  assumes ok: "isOK (check_subsumptions_guided E convs)"
  shows "subsumes (set E) (fst ` set convs)"
    unfolding subsumes_via_rule_conversion
    using ok
proof (induct convs arbitrary: E)
  case Nil show ?case by simp
next
  case (Cons ec convs)
  obtain s t conv where ec: "ec = ((s,t),conv)" by (cases ec, force)
  note Cons = Cons[unfolded ec]
  from Cons(2) have rec: "isOK(check_subsumptions_guided ((s,t) # E) convs)" 
    and one: "isOK(check_conversion E conv s t)" by auto
  from check_conversion_sound[OF one] have one: "(s,t) \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*" .
  {
    fix l r
    assume "(l,r) \<in> fst ` set convs"
    with Cons(1)[OF rec] have "(l,r) \<in> (rstep (set ((s,t) # E)))\<^sup>\<leftrightarrow>\<^sup>*" by auto
    then have "(l,r) \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def
    proof (induct)
      case (step r u)
      from step(2) have "(r,u) \<in> (rstep {(s,t)})\<^sup>\<leftrightarrow> \<or> (r,u) \<in> (rstep (set E))\<^sup>\<leftrightarrow>"
        unfolding estep_sym_closure_conv using rstep_union[of "{(s,t)}^<->" "(set E)^<->"]
        by auto
      then have "(r,u) \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*"
      proof
        assume step: "(r,u) \<in> (rstep {(s,t)})\<^sup>\<leftrightarrow>"
        have "subsumes (set E) {(s,t)}"
          unfolding subsumes_via_rule_conversion using one by auto
        then show ?thesis unfolding subsumes_def eq_theory_is_esteps
          using step unfolding conversion_def by auto
      qed auto
      with step(3) show ?case by auto
    qed auto
  }
  then show ?case unfolding ec using one by force
qed

definition check_subsumption_guided :: "('f :: showl, 'v :: showl) equation list \<Rightarrow> ('f,'v)equation list \<Rightarrow> ('f, 'v) subsumption_proof \<Rightarrow> showsl check"
where  "check_subsumption_guided E E' convs \<equiv> do {
    check_subseteq E (map fst convs)
      <+? (\<lambda>r. showsl_lit (STR ''could not find conversion for equation '') \<circ> showsl_eq r);
    check_subsumptions_guided E' convs
  }"

lemma check_subsumption_guided:
  assumes ok: "isOK (check_subsumption_guided E E' convs)"
  shows "subsumes (set E') (set E)"
proof -
  note ok = ok[unfolded check_subsumption_guided_def]
  from ok have subset: "set E \<subseteq> fst ` set convs" and
    ok: "isOK (check_subsumptions_guided E' convs)" by auto
  from check_subsumptions_guided[OF ok] subset 
  show ?thesis  
    unfolding subsumes_via_rule_conversion by blast
qed

definition check_single_subsumption :: "('f :: showl, 'v :: showl) equation \<Rightarrow> ('f,'v)equation list \<Rightarrow> ('f, 'v) subsumption_proof \<Rightarrow> showsl check"
where  "check_single_subsumption eq E convs \<equiv> do {
    check (eq \<in> set (map fst convs)) (showsl_lit (STR ''could not find conversion for equation '') \<circ> showsl_eq eq);
    check_subsumptions_guided E convs
  }"

lemma check_single_subsumption:
  assumes ok: "isOK (check_single_subsumption eq E convs)"
  shows "eq \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  note ok = ok[unfolded check_single_subsumption_def]
  from ok have mem: "eq \<in> fst ` set convs" and
    ok: "isOK (check_subsumptions_guided E convs)" by auto
  from check_subsumptions_guided[OF ok] mem
  show ?thesis  
    unfolding subsumes_via_rule_conversion by force
qed

definition check_convertible_instance :: "('f :: showl, 'v :: showl) equation \<Rightarrow> ('f,'v)equation list \<Rightarrow> ('f, 'v) subsumption_proof \<Rightarrow> showsl check"
where  "check_convertible_instance eq E convs \<equiv> do {
    check_exm (\<lambda>c. check (instance_rule (fst c) eq) (showsl_lit (STR ''not an instance of '') \<circ> showsl_eq eq)) convs (\<lambda>_. showsl_lit (STR '' no instance found''));
    check_subsumptions_guided E convs
  }"

lemma check_convertible_instance:
  assumes ok: "isOK (check_convertible_instance (s,t) E convs)"
  shows "\<exists>\<sigma>. (s\<cdot>\<sigma>, t\<cdot>\<sigma>) \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  note ok = ok[unfolded check_convertible_instance_def]
  from ok obtain c where inst: "fst c \<in> fst ` set convs" "instance_rule (fst c) (s,t)"
    and ok: "isOK (check_subsumptions_guided E convs)" by fastforce
  obtain u v where uv:"fst c = (u,v)" by fastforce
  from inst obtain \<sigma> where \<sigma>: "u = s \<cdot> \<sigma> \<and> v = t \<cdot> \<sigma>" unfolding uv instance_rule_def by auto
  note sub = check_subsumptions_guided[OF ok, unfolded subsumes_via_rule_conversion, rule_format, of "fst c"]
  with \<sigma> inst(1) show ?thesis unfolding uv by force
qed

lemma check_join_NF_conversion: 
  assumes ok: "isOK (check_join_NF R s t)"
  shows "(s,t) \<in> (rstep (set R))\<^sup>\<leftrightarrow>\<^sup>*"
  by (rule set_mp[OF join_imp_conversion check_join_NF_sound[OF ok]])

definition
  check_subsumption_NF :: "('f:: showl, 'v:: showl) equation list \<Rightarrow> ('f, 'v) rules \<Rightarrow> showsl check"
where
  "check_subsumption_NF E R \<equiv>  do {
    check_allm (\<lambda> e. check_join_NF R (fst e) (snd e)
      <+? (\<lambda>s. showsl_lit (STR ''could not join equation '') \<circ> showsl_eq e \<circ> showsl_nl \<circ> s)) E
  }"

lemma check_subsumption_NF:
  assumes ok: "isOK (check_subsumption_NF E R)"
  shows "subsumes (set R) (set E)"
proof -
  note ok = ok[unfolded check_subsumption_NF_def, simplified]
  show ?thesis
    unfolding subsumes_via_rule_conversion
  proof (clarify)
    fix s t
    assume "(s,t) \<in> set E"
    with ok have "isOK (check_join_NF R s t)" by auto
    from check_join_NF_conversion[OF this]
    show "(s, t) \<in> (rstep (set R))\<^sup>\<leftrightarrow>\<^sup>*" .
  qed
qed

definition
  check_subsumption :: "('f:: showl, 'v:: showl) equation list \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) subsumption_proof option \<Rightarrow> showsl check"
where
  "check_subsumption E R convs_o \<equiv>
  (case convs_o of
    None \<Rightarrow> check_subsumption_NF E R
  | Some convs \<Rightarrow> check_subsumption_guided E R convs)"

lemma check_subsumption:
  assumes ok: "isOK (check_subsumption E R convs_o)"
  shows "subsumes (set R) (set E)"
proof -
  note ok = ok[unfolded check_subsumption_def]
  show ?thesis
  proof (cases convs_o)
    case None
    show ?thesis 
      by (rule check_subsumption_NF, insert None ok, auto)
  next
    case (Some convs)
    show ?thesis 
      by (rule check_subsumption_guided, insert Some ok, auto)
  qed
qed

end
