(*
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
License: LGPL (see file COPYING.LESSER)
*)
theory WPO_Impl
  imports
    Weighted_Path_Order.WPO
    Term_Order_Impl
    Status_Impl
    TRS.Term_Impl
    Auxx.Name
    Auxx.Shows_Literal_List
begin


definition af_compatible_status :: "'f af \<Rightarrow> 'f status \<Rightarrow> bool"
where
  "af_compatible_status \<pi> \<sigma> \<longleftrightarrow> (\<forall> f m. set (status \<sigma> (f, m)) \<subseteq> \<pi> (f, m))"

lemma af_compatible_status_full_af: "af_compatible_status full_af \<sigma>"
  unfolding af_compatible_status_def full_af_def using status[of \<sigma>] by auto

locale wpo_params = ws_redpair_order S NS \<sigma>\<sigma> + precedence prc prl
  for NS S :: "('f, 'v) trs" 
    and prc :: "'f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool"
    and prl :: "'f \<times> nat \<Rightarrow> bool"
    and ssimple :: bool
    and large :: "'f \<times> nat \<Rightarrow> bool" 
    and n :: nat
    and \<sigma>\<sigma> :: "'f status" +
  assumes large: "ssimple \<Longrightarrow> large fn \<Longrightarrow> fst (prc fn gm) \<or> snd (prc fn gm) \<and> status \<sigma>\<sigma> gm = []"  
  and large_trans: "ssimple \<Longrightarrow> large fn \<Longrightarrow> snd (prc gm fn) \<Longrightarrow> large gm"  
  and strictly_simple: "ssimple \<Longrightarrow> ss_redpair_order S NS \<sigma>\<sigma>" 
begin

sublocale wpo_with_assms 
proof (unfold_locales; (intro S_imp_NS large ws_status S_stable NS_stable NS_mono; assumption)?)
  assume ssimple
  show "large fn \<Longrightarrow> snd (prc gm fn) \<Longrightarrow> large gm" for fn gm by (rule large_trans[OF `ssimple`])
  interpret ss_redpair_order S NS \<sigma>\<sigma> by (rule strictly_simple[OF `ssimple`])
  show "S \<noteq> {}" by (rule S_non_empty)
  show "i \<in> set (status \<sigma>\<sigma> fn) \<Longrightarrow> simple_arg_pos S fn i " for fn i by (rule ss_status)
qed


theorem wpo_redtriple_order: "redtriple_order WPO_S WPO_NS WPO_NS"
proof -
  interpret SN_order_pair WPO_S WPO_NS by (rule WPO_SN_order_pair)
  show ?thesis
  proof (unfold_locales; (intro trans_NS refl_NS compat_NS_S WPO_S_subset_WPO_NS)?)
    show "subst.closed WPO_S" using wpo_stable by auto
    show "subst.closed WPO_NS" using wpo_stable by auto
    show "ctxt.closed WPO_NS" using wpo_ns_mono one_imp_ctxt_closed[of WPO_NS] by blast
  qed
qed


lemma ce_\<sigma>: assumes "{0,1} \<subseteq> set (\<sigma> (f,Suc (Suc k)))"
  shows "ce_trs (f,k) \<subseteq> WPO_S \<and> ce_trs (f,k) \<subseteq> WPO_NS"
proof -
  {
    fix s t :: "('f,'v)term"
    assume "(s,t) \<in> ce_trs (f,k)"
    from this[unfolded ce_trs.simps] 
    obtain u v where s: "s = Fun f (u # v # replicate k (Var undefined))" (is "_ = Fun _ ?ss")
      and t: "t = u \<or> t = v" by auto
    from t subterm_wpo_s_arg[of 0 f ?ss] subterm_wpo_s_arg[of 1 f ?ss] assms 
    have "(s,t) \<in> WPO_S" unfolding s by (cases, auto)
  }
  then show ?thesis using wpo_s_imp_ns by blast
qed

context
  fixes \<pi> :: "'f af"
  assumes af_NS: "af_compatible \<pi> NS"
  and af_\<sigma>: "af_compatible_status \<pi> \<sigma>\<sigma>"
begin

lemma af_\<sigma>: "i < m \<Longrightarrow> i \<in> \<pi> (f,m) \<or> i \<notin> set (\<sigma> (f,m))"
  using af_\<sigma>[unfolded af_compatible_status_def] by blast

lemma wpo_af_compatible: "af_compatible \<pi> WPO_NS"
  unfolding af_compatible_def
proof (intro allI)
  fix f and bef aft :: "('f,'v)term list" and s t
  let ?i = "length bef"
  let ?n = "Suc (?i + length aft)"
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  let ?t = "Fun f ?ts"
  let ?\<sigma> = "\<sigma> (f,?n)"
  show "?i \<in> \<pi> (f, ?n) \<or> (?s, ?t) \<in> WPO_NS" 
  proof (cases "?i \<in> \<pi> (f, ?n)")
    case False
    have i: "?i < ?n" by auto
    from af_\<sigma>[OF i, of f] False have i\<sigma>: "?i \<notin> set ?\<sigma>" by auto
    from af_NS[unfolded af_compatible_def, rule_format, of bef f aft s t] False
    have id: "((?s, ?t) \<in> NS) = True" "length ?ss = ?n" "length ?ts = ?n" by auto
    have "\<forall>j\<in>set ?\<sigma>. wpo_s ?s (?ts ! j)" (is ?one)
    proof
      fix j
      assume j: "j \<in> set ?\<sigma>"
      with i\<sigma> have ji: "j \<noteq> ?i" by auto
      then have id: "?ts ! j = ?ss ! j" by (rule append_Cons_nth_not_middle)      
      show "wpo_s ?s (?ts ! j)" unfolding id 
        by (rule subterm_wpo_s_arg, insert j, auto)
    qed
    then have one: "?one = True" by simp
    have two: "map ((!) ?ts) ?\<sigma> = map ((!) ?ss) ?\<sigma>" (is "?mts = ?mss")
    proof (rule nth_equalityI, force, unfold length_map)
      fix i
      assume i: "i < length ?\<sigma>"
      with i\<sigma> have "?\<sigma> ! i \<noteq> ?i" unfolding set_conv_nth by auto
      then show "?mts ! i = ?mss ! i" using i by (simp add: nth_append)
    qed
    have "snd (lex_ext wpo_pr n ?mss ?mts)" unfolding two
      by (rule all_nstri_imp_lex_nstri, insert wpo_ns_refl, auto)
    then have "wpo_ns ?s ?t" unfolding wpo_simps[of ?s ?t] term.simps id one
      prc_refl  by auto
    then show ?thesis by blast
  qed simp
qed
end

lemma wpo_ce_af_redtriple_order: 
  assumes ce: "\<exists>n. \<forall>m\<ge>n. \<forall> f. {0,1} \<subseteq> set (\<sigma> (f, Suc (Suc m)))"
  and pi: "af_compatible \<pi> NS" "af_compatible_status \<pi> \<sigma>\<sigma>"
  shows "mono_ce_af_redtriple_order WPO_S WPO_NS WPO_NS \<pi>"
proof -
  interpret redtriple_order WPO_S WPO_NS WPO_NS by (rule wpo_redtriple_order)
  from assms obtain n where "\<And> m f. m \<ge> n \<Longrightarrow> {0,1} \<subseteq> set (\<sigma> (f, Suc (Suc m)))" by blast
  from ce_\<sigma>[OF this] have "ce_compatible WPO_NS" "ce_compatible WPO_S" 
    unfolding ce_compatible_def by blast+ 
  with wpo_af_compatible[OF pi]
  show ?thesis using full_af by (unfold_locales, blast+)
qed 
    
  
context
  assumes \<sigma>_full: "\<And> f k. set (\<sigma> (f,k)) = {0 ..< k}"
begin

lemma ctxt_closed_WPO_S: "ctxt.closed WPO_S" 
  using one_imp_ctxt_closed[OF WPO_S_ctxt[OF \<sigma>_full]] by blast

lemma wpo_mono_ce_af_redtriple_order: 
  "mono_ce_af_redtriple_order WPO_S WPO_NS WPO_NS full_af"
  by (rule wpo_ce_af_redtriple_order, 
    insert \<sigma>_full full_af af_compatible_status_full_af, auto)
end
end


(*TODO: move*)
\<comment> \<open>show function symbol together with arity\<close>
fun showsl_funa :: "('f::showl \<times> nat) \<Rightarrow> showsl"
  where
    "showsl_funa (f, n) = showsl f \<circ> showsl_lit (STR ''/'') \<circ> showsl n"

definition compare_bools :: "bool \<times> bool \<Rightarrow> bool \<times> bool \<Rightarrow> bool"
  where
    "compare_bools p1 p2 \<longleftrightarrow> (fst p1 \<longrightarrow> fst p2) \<and> (snd p1 \<longrightarrow> snd p2)"

notation compare_bools ("(_/ \<le>\<^sub>c\<^sub>b _)" [51, 51] 50)

lemma lex_ext_unbounded_cb:
  assumes "\<And> i. i < length xs \<Longrightarrow> i < length ys \<Longrightarrow> f (xs ! i) (ys ! i) \<le>\<^sub>c\<^sub>b g (xs ! i) (ys ! i)"
  shows "lex_ext_unbounded f xs ys \<le>\<^sub>c\<^sub>b lex_ext_unbounded g xs ys"
  unfolding compare_bools_def
  by (rule lex_ext_unbounded_mono,
  insert assms[unfolded compare_bools_def], auto)

context
  fixes "pr" :: "('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool)"
    and prl :: "'f \<times> nat \<Rightarrow> bool"
    and ssimple :: bool
    and large :: "'f \<times> nat \<Rightarrow> bool"
    and cS cNS :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
    and \<sigma> :: "'f status"
begin

fun wpo_ub  :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool" 
  where
    "wpo_ub s t = (case s of
      Var x \<Rightarrow> (False,
        (case t of
          Var y \<Rightarrow> x = y
        | Fun g ts \<Rightarrow> cNS s t \<and> status \<sigma> (g, length ts) = [] \<and> prl (g, length ts)))
    | Fun f ss \<Rightarrow>
      if cS s t then (True, True)
      else
        let ff = (f, length ss); sf = status \<sigma> ff in
        if cNS s t then
          if (\<exists> i \<in> set sf. snd (wpo_ub (ss ! i) t)) then (True, True)
          else
            (case t of
              Var _ \<Rightarrow> (False, ssimple \<and> large ff)
            | Fun g ts \<Rightarrow>
              let gg = (g, length ts); sg = status \<sigma> gg in
              (case pr ff gg of (prs, prns) \<Rightarrow>
                if prns \<and> (\<forall> j \<in> set sg. fst (wpo_ub s (ts ! j))) then
                  if prs then (True, True)
                  else lex_ext_unbounded wpo_ub (map (\<lambda> i. ss ! i) sf) (map (\<lambda> i. ts ! i) sg)
                else (False, False)))
        else (False, False))"

declare wpo_ub.simps [simp del]

abbreviation "wpo_orig \<equiv> wpo pr prl ssimple large"

lemma wpo_ub:
  assumes "\<And> si tj. s \<unrhd> si \<Longrightarrow> t \<unrhd> tj \<Longrightarrow> (cS si tj, cNS si tj) \<le>\<^sub>c\<^sub>b ((si, tj) \<in> S, (si, tj) \<in> NS)"
    and "\<And> f. f \<in> funas_term t \<Longrightarrow> length (status \<sigma> f) \<le> n"
  shows "wpo_ub s t \<le>\<^sub>c\<^sub>b wpo_orig n S NS \<sigma> s t" 
  using assms
proof (induct s t rule: wpo.induct [of S NS \<sigma> _ "pr" prl ssimple large n])
  case (1 s t) 
  note IH = 1(1-3)
  note cb = 1(4)
  note n = 1(5)
  note cbd = compare_bools_def
  note simps = wpo_ub.simps[of s t] wpo.simps[of "pr" prl ssimple large n S NS \<sigma> s t]
  let ?wpo = "wpo_orig n S NS \<sigma>"
  let ?cb = "\<lambda> s t. (cS s t, cNS s t) \<le>\<^sub>c\<^sub>b ((s, t) \<in> S, (s, t) \<in> NS)"
  let ?goal = "\<lambda> s t. wpo_ub s t \<le>\<^sub>c\<^sub>b ?wpo s t"
  from cb[of s t] have cb_st: "?cb s t" by auto
  show ?case
  proof (cases s)
    case (Var x) note s = this
    show ?thesis
    proof (cases t)
      case (Var y) note t = this
      show ?thesis unfolding simps unfolding s t cbd by simp
    next
      case (Fun g ts) note t = this
      show ?thesis unfolding simps using cb_st unfolding s t cbd by auto
    qed
  next
    case (Fun f ss) note s = this
    let ?f = "(f,length ss)"
    let ?sf = "status \<sigma> ?f"
    let ?s = "Fun f ss"
    show ?thesis
    proof (cases "(s,t) \<in> S \<or> \<not> cNS s t")
      case True
      with cb_st show ?thesis unfolding simps unfolding s cbd by auto
    next
      case False
      with cb_st have *: "(s,t) \<notin> S" "(s,t) \<in> NS" "((s,t) \<notin> S) = True" "((s,t) \<in> NS) = True" "cS s t = False" "cNS s t = True"
        unfolding cbd by auto
      note IH = IH[OF s *(1-2)]
      show ?thesis
      proof (cases "(\<exists> i \<in> set ?sf. snd (?wpo (ss ! i) t))")
        case True
        then show ?thesis unfolding simps using True * unfolding s cbd by auto
      next
        case False
        {
          fix i
          assume i: "i \<in> set ?sf"
          from status_aux[OF i] 
          have "?goal (ss ! i) t"
            by (intro IH(1)[OF i cb n], auto simp: s)
        }
        with False have sub: "(\<exists> i \<in> set ?sf. snd (wpo_ub (ss ! i) t)) = False" unfolding cbd by auto
        note IH = IH(2-3)[OF False]
        show ?thesis
        proof (cases "wpo_ub s t = (False,False)")
          case True
          then show ?thesis unfolding cbd by auto
        next
          case False note noFF = this
          note False = False[unfolded simps * Let_def, unfolded s term.simps sub, simplified]
          show ?thesis
          proof (cases t)
            case t: (Var y)
            from False[unfolded t, simplified]
            show ?thesis unfolding s t unfolding cbd
              using * Fun simps sub t by auto
          next
            case t: (Fun g ts)
            let ?g = "(g,length ts)"
            let ?sg = "status \<sigma> ?g"
            let ?t = "Fun g ts"
            obtain ps pns where p: "pr ?f ?g = (ps, pns)" by force
            note IH = IH[OF t p[symmetric]]
            note False = False[unfolded t p split term.simps]
            from False have pns: "pns = True" by (cases pns, auto)
            {
              fix j
              assume j: "j \<in> set ?sg"
              from status_aux[OF j] 
              have cb: "?goal s (ts ! j)"
                by (intro IH(1)[OF j cb n], auto simp: t)
              from j False have "fst (wpo_ub s (ts ! j))" unfolding s by (auto split: if_splits)
              with cb have "fst (?wpo s (ts ! j))" unfolding cbd by auto
            }
            then have cond: "pns \<and> (\<forall> j \<in> set ?sg. fst (?wpo s (ts ! j)))" using pns by auto
            note IH = IH(2)[OF cond]
            from cond have cond: "(pns \<and> (\<forall> j \<in> set ?sg. fst (?wpo ?s (ts ! j)))) = True" unfolding s by simp
            note simps = simps[unfolded * Let_def, unfolded s t term.simps if_False if_True sub[unfolded t] p split cond]
            show ?thesis 
            proof (cases ps)
              case True
              then show ?thesis unfolding s t unfolding simps cbd by auto
            next
              case False
              note IH = IH[OF this _ _ cb n, unfolded s t length_map]
              let ?msf = "map ((!) ss) ?sf"
              let ?msg = "map ((!) ts) ?sg"
              from n[of ?g, unfolded t] have "length (?msg) \<le> n" by auto
              then have ub: "lex_ext_unbounded ?wpo ?msf ?msg =
              lex_ext ?wpo n ?msf ?msg" 
                unfolding lex_ext_unbounded_iff lex_ext_iff by auto
              {
                fix i
                assume "i < length ?msf"
                then have "?msf ! i \<in> set ?msf" unfolding set_conv_nth by blast
                also have "set ?msf \<subseteq> set ss" using status[of \<sigma> f "length ss"]
                  unfolding set_conv_nth by force
                finally have "?msf ! i \<in> set ss" .
              } note msf = this
              {
                fix i
                assume "i < length ?msg"
                then have "?msg ! i \<in> set ?msg" unfolding set_conv_nth by blast
                also have "set ?msg \<subseteq> set ts" using status[of \<sigma> g "length ts"]
                  unfolding set_conv_nth by force
                finally have "?msg ! i \<in> set ts" .
              } note msg = this
              from False simps noFF
              have "wpo_ub s t = lex_ext_unbounded wpo_ub ?msf ?msg"
                unfolding s t simps by (auto split: if_splits)
              also have "\<dots> \<le>\<^sub>c\<^sub>b lex_ext_unbounded ?wpo ?msf ?msg"
                by (rule lex_ext_unbounded_cb, rule IH, insert msf msg, auto)
              finally show ?thesis unfolding ub s t simps(2) cbd by auto
            qed
          qed
        qed
      qed
    qed
  qed
qed

end

definition af_wpo :: "'f af \<Rightarrow> 'f status \<Rightarrow> 'f af"
  where
    "af_wpo \<pi> \<sigma> f = set (status \<sigma> f) \<union> \<pi> f"

lemma af_wpo_compat: "af_compatible \<pi> NS \<Longrightarrow> af_compatible (af_wpo \<pi> \<sigma>) NS"
  unfolding af_compatible_def af_wpo_def by auto

lemma af_wpo_status: "af_compatible_status (af_wpo \<pi> \<sigma>) \<sigma>"
  unfolding af_compatible_status_def af_wpo_def by auto

definition faulty_redtriple :: "'f itself \<Rightarrow> 'v itself \<Rightarrow> showsl \<Rightarrow> showsl \<Rightarrow> ('f, 'v)redtriple"
  where
    "faulty_redtriple _ _ err desc = \<lparr>
      redtriple.valid = error err,
      s = (\<lambda> _. succeed),
      ns = (\<lambda> _. succeed),
      nst = (\<lambda> _. succeed),
      af = full_af,
      mono_af = empty_af,
      mono = (\<lambda> _ . succeed),
      desc = desc,
      not_wst = None,
      not_sst = None,
      cpx = no_complexity_check
    \<rparr>"

lemma faulty_redtriple [simp]: "\<not> isOK (redtriple.valid (faulty_redtriple a b c d))"
  unfolding faulty_redtriple_def by simp

abbreviation var_x_i :: "nat => ('f, string) term"
  where
    "var_x_i i == Var (''x'' @ show i)"

fun check_status_ws_info :: "'f status \<Rightarrow> (('f, string) rule \<Rightarrow> showsl check) \<Rightarrow> (('f :: showl) \<times> nat) list option \<Rightarrow> showsl check"
  where
    "check_status_ws_info \<sigma> cns None = error (showsl_lit (STR ''missing weak-subterm status of base reduction pair''))"
  | "check_status_ws_info \<sigma> cns (Some fs) = check_allm (\<lambda>(f, n).
      check_allm (\<lambda>i.
        let s = Fun f (map var_x_i [0..<n]); t = var_x_i i in
        cns (s, t)
        <+? (\<lambda>_. showsl_lit (STR ''argument #'') \<circ> showsl (Suc i) \<circ>
          showsl_lit (STR '' is in status of '') \<circ> showsl_funa (f, n) \<circ> showsl_nl \<circ>
          showsl_lit (STR ''but '') \<circ> showsl s \<circ> showsl_lit (STR '' >= '') \<circ> showsl_nat_var i \<circ>
          showsl_lit (STR '' is not satisfied''))) (status \<sigma> (f, n))) fs"

type_synonym 'f wpo_params = "(('f \<times> nat) \<times> (nat \<times> nat list)) list"
(* symbol * prec * status *)

definition showsl_wpo_params :: "('f :: showl) wpo_params \<Rightarrow> showsl" where
  "showsl_wpo_params params = showsl_lit (STR ''status and precedence:\<newline>'') \<circ>
    showsl_sep (\<lambda>(f, (p, s)).
      showsl_lit (STR ''precedence('') \<circ> showsl_funa f \<circ> showsl_lit (STR '') = '') \<circ> showsl p \<circ> showsl_nl \<circ>
      showsl_lit (STR ''  status('') \<circ> showsl_funa f \<circ> showsl_lit (STR '') = '') \<circ> showsl_list s
    ) showsl_nl params"

definition large_of where 
  "large_of pr \<sigma> fs = (let m = max_list (map pr fs);
     ls = filter (\<lambda> f. pr f = m) fs
     in if m > 0 \<and> (\<forall> f \<in> set ls. status \<sigma> f = []) then Some m else None)" 

definition wpo_redtriple :: "('f :: {compare_order, showl}, string) redtriple \<Rightarrow> 'f wpo_params \<Rightarrow> ('f, string) redtriple"
  where
    "wpo_redtriple rt params \<equiv>
      (let
        stat = map (\<lambda> (f,ps). (f, snd ps)) params;
        pr = fun_of_map_fun' (ceta_map_of params) (\<lambda> _. 0) fst;
        desc1 = showsl_lit (STR ''WPO '');
        desc2 = showsl_lit (STR ''with '') o showsl_wpo_params params \<circ>
          showsl_lit (STR ''\<newline>over the following reduction pair:\<newline>'') \<circ> redtriple.desc rt
      in
        (case status_of stat of
          None \<Rightarrow> faulty_redtriple (TYPE('f)) (TYPE(string)) (showsl_lit (STR ''problem with indices in status of WPO!'')) (desc1 o desc2)
        | Some \<sigma> \<Rightarrow>
          let
            large_opt = large_of pr \<sigma> (map fst params);
            ssimple = \<not> is_None large_opt \<and> isOK(check_status_ws_info \<sigma> (redtriple.s rt) (redtriple.not_sst rt));
            large = (if ssimple then \<lambda> f. pr f = the large_opt else (\<lambda> _. False));
            s = (\<lambda> s t. isOK (redtriple.s rt (s, t)));
            ns = (\<lambda> s t. isOK (redtriple.ns rt (s, t)));
            wpo = wpo_ub (prc_nat pr) (prl_nat pr) ssimple large s ns \<sigma>;
            wpo_s = (\<lambda> (s,t). check (fst (wpo s t)) (showsl s \<circ> showsl_lit (STR '' >wpo '') \<circ> showsl t \<circ> showsl_lit (STR '' could not be ensured'')));
            wpo_ns = (\<lambda> (s,t). check (snd (wpo s t)) (showsl s \<circ> showsl_lit (STR '' >=wpo '') \<circ> showsl t \<circ> showsl_lit (STR '' could not be ensured'')))
          in \<lparr>
            redtriple.valid = do { 
              redtriple.valid rt; 
              check_status_ws_info \<sigma> (redtriple.ns rt) (redtriple.not_wst rt) 
            },
            s = wpo_s,
            ns = wpo_ns,
            nst = wpo_ns,
            af = af_wpo (redtriple.af rt) \<sigma>,
            mono_af = empty_af, \<comment> \<open>TODO: this is too crude\<close>
            mono = (\<lambda>_. check_allm (\<lambda>((f, n), idx). check (set idx = set [0..<n]) (
              showsl_lit (STR ''for monotonicity, status must be complete, but status of '') \<circ>
              showsl_funa (f, n) \<circ> showsl_lit (STR '' is '') \<circ> showsl (map Suc idx))) stat),
            desc = if ssimple then desc1 o showsl_lit (STR ''(strictly simple) '') o desc2 else desc1 o desc2,
            not_wst = Some (map fst stat),
            not_sst = Some (map fst stat),
            cpx = no_complexity_check
          \<rparr>))"

lemma wpo_redtriple: assumes gen: "\<And> s ns nst. isOK(redtriple.valid rt) \<Longrightarrow> 
  isOK (try forallM (redtriple.s rt) s catch (\<lambda>x. Inl (snd x))) \<Longrightarrow>
  isOK (try forallM (redtriple.ns rt) ns catch (\<lambda>x. Inl (snd x))) \<Longrightarrow>
  isOK (try forallM (redtriple.nst rt) nst catch (\<lambda>x. Inl (snd x))) \<Longrightarrow>
  \<exists>S NS NST.
     cpx_ce_af_redtriple_order S NS NST (redtriple.af rt) (redtriple.mono_af rt) (\<lambda> cm cc. isOK(redtriple.cpx rt cm cc)) \<and>
     set s \<subseteq> S \<and>
     set ns \<subseteq> NS \<and>
     set nst \<subseteq> NST \<and>
     not_subterm_rel_info NS (redtriple.not_wst rt) \<and>
     not_subterm_rel_info S (redtriple.not_sst rt) \<and>
     (isOK (redtriple.mono rt (s @ ns @ nst)) \<longrightarrow>
      mono_ce_af_redtriple_order S NS NST (redtriple.af rt) \<and> ctxt.closed S)"
  shows "generic_redtriple_impl (wpo_redtriple rt)"
proof
  fix param and s_list ns_list nst_list 
  let ?rp = "wpo_redtriple rt param"
  let ?pi = "redtriple.af ?rp"
  let ?mpi = "redtriple.mono_af ?rp"
  let ?cpx = "redtriple.cpx ?rp"
  let ?Cpx = "\<lambda> cm cc. isOK(?cpx cm cc)"
  let ?s = "\<lambda> s t. isOK(redtriple.s ?rp (s,t))"
  let ?ns = "\<lambda> s t. isOK(redtriple.ns ?rp (s,t))"
  let ?stat = "map (\<lambda> (f,ps). (f, snd ps)) param"
  assume valid: "isOK(redtriple.valid ?rp)" and stri: "isOK (check_allm (redtriple.s ?rp) s_list)" 
    and nstri: "isOK (check_allm (redtriple.ns ?rp) ns_list)"
    and nstrit: "isOK (check_allm (redtriple.nst ?rp) nst_list)"
  let ?rp' = "rt"
  let ?pi' = "redtriple.af ?rp'"
  let ?mpi' = "redtriple.mono_af ?rp'"
  let ?cpx' = "redtriple.cpx ?rp'"
  let ?Cpx' = "\<lambda> cm cc. isOK(?cpx' cm cc)"
  let ?s' = "\<lambda> s t. isOK(redtriple.s ?rp' (s,t))"
  let ?ns' = "\<lambda> s t. isOK(redtriple.ns ?rp' (s,t))"
  let ?pr = "fun_of_map_fun' (ceta_map_of param) (\<lambda> _. 0) fst"
  let ?prc = "prc_nat ?pr"
  let ?prl = "prl_nat ?pr"
  let ?ws' = "redtriple.not_wst ?rp'"
  let ?ss' = "redtriple.not_sst ?rp'"
  note d = wpo_redtriple_def[unfolded Let_def]
  note valid = valid[unfolded d param split]
  from valid obtain \<sigma> where stat: "status_of ?stat = Some \<sigma>" by (cases "status_of ?stat", auto)
  let ?large_opt = "large_of ?pr \<sigma> (map fst param)" 
  define ssimple where "ssimple = (\<not> is_None ?large_opt \<and> isOK (check_status_ws_info \<sigma> (redtriple.s rt) (redtriple.not_sst rt)))" 
  define large where "large = (if ssimple then \<lambda> f. ?pr f = (the ?large_opt) else (\<lambda> _ :: ('a \<times> nat). False))" 
  note param = param stat
  note valid = valid[unfolded stat option.simps, folded ssimple_def, folded large_def]
  note stri = stri[unfolded d param split option.simps, folded ssimple_def, folded large_def]
  note nstri = nstri[unfolded d param split option.simps, folded ssimple_def, folded large_def]
  note nstrit = nstrit[unfolded d param split option.simps, folded ssimple_def, folded large_def]
  let ?wpo = "wpo_ub (prc_nat ?pr) (prl_nat ?pr) ssimple large ?s' ?ns' \<sigma>"
  from valid have status_ws: "isOK(check_status_ws_info \<sigma> (redtriple.ns ?rp') ?ws')" 
    and "isOK(redtriple.valid ?rp')" by auto
  note gen = gen[OF this(2)]
  from status_ws obtain fs where ws': "?ws' = Some fs" by (cases ?ws', auto)
  from status_ws[unfolded ws']
  have status_ws: "\<And>f n i. (f, n) \<in> set fs \<Longrightarrow> i \<in> set (status \<sigma> (f, n)) \<Longrightarrow>
    ?ns' (Fun f (map var_x_i [0..<n])) (var_x_i i)" by auto
  have status_ss: "\<And>f n i. ssimple \<Longrightarrow> (f, n) \<in> set (the ?ss') \<Longrightarrow> i \<in> set (status \<sigma> (f, n)) \<Longrightarrow>
    ?s' (Fun f (map var_x_i [0..<n])) (var_x_i i)" unfolding ssimple_def by (cases ?ss', auto)
  define all where "all = s_list @ ns_list @ nst_list"
  define subts where "subts = [(u,v) . (s,t) <- all, u <- supteq_list s, v <- supteq_list t]"
  define ssimple_comps where "ssimple_comps = (if ssimple then
    [(Fun f (map var_x_i [0..<n]), var_x_i i :: ('a, _) Term.term). (f, n) <- the ?ss', i <- status \<sigma> (f, n)] else [])" 
  define s where "s = filter (\<lambda> (s,t). ?s' s t) subts @ ssimple_comps" 
  define ns where "ns = (filter (\<lambda> (s,t). ?ns' s t) subts) 
    @ [(Fun f (map var_x_i [0..<n]), var_x_i i). (f, n) <- fs, i <- status \<sigma> (f, n)]"
  have "\<exists>S NS NST.
     cpx_ce_af_redtriple_order S NS NST ?pi' ?mpi' ?Cpx' \<and>
     set s \<subseteq> S \<and> set ns \<subseteq> NS \<and> set [] \<subseteq> NST \<and>
     not_subterm_rel_info NS ?ws' \<and>
     not_subterm_rel_info S ?ss' \<and>
     (isOK (redtriple.mono rt (s @ ns @ [])) \<longrightarrow>
        mono_ce_af_redtriple_order S NS NST ?pi' \<and> ctxt.closed S)"
    by (rule gen, insert status_ws status_ss, auto simp: o_def s_def ns_def ssimple_comps_def)
  then obtain S NS NST where
    rt: "cpx_ce_af_redtriple_order S NS NST ?pi' ?mpi' ?Cpx'"
    and orient: "set s \<subseteq> S \<and> set ns \<subseteq> NS" 
    and ws: "not_subterm_rel_info NS ?ws'" 
    and sst: "not_subterm_rel_info S ?ss'" by blast
  define n where "n = max_list [ length (status \<sigma> f) . (s,t) <- all, f <- funas_term_list t]"
  interpret cpx_ce_af_redtriple_order S NS NST ?pi' ?mpi' ?Cpx' by fact
  interpret precedence ?prc ?prl ..
  note cb = compare_bools_def
  let ?wpoo = "wpo_orig ?prc ?prl ssimple large n S NS \<sigma>"
  let ?wpo_s = "{(s, t). fst (?wpoo s t)}"
  let ?wpo_ns = "{(s, t). snd (?wpoo s t)}"
  {
    fix s t
    assume st: "(s,t) \<in> set all"
    have "?wpo s t \<le>\<^sub>c\<^sub>b ?wpoo s t"
    proof (rule wpo_ub)
      fix si tj
      assume "s \<unrhd> si" "t \<unrhd> tj"
      with st have sitj: "(si,tj) \<in> set subts" unfolding subts_def by force
      with orient
      show "(?s' si tj, ?ns' si tj)
       \<le>\<^sub>c\<^sub>b ((si, tj) \<in> S, (si, tj) \<in> NS)" unfolding s_def ns_def cb by auto
    next
      fix f
      assume f: "f \<in> funas_term t"
      show "length (status \<sigma> f) \<le> n" unfolding n_def
        by (rule max_list, insert f st, auto)
    qed
  } note cb_all = this[unfolded cb]
  interpret wpo_params NS S ?prc ?prl ssimple large n \<sigma> 
  proof
    show "S \<subseteq> NS" "S \<subseteq> NS" by (rule S_imp_NS)+
    fix i fn 
    assume i: "i \<in> set (status \<sigma> fn)"
    obtain f n where f: "fn = (f,n)" by force
    show "simple_arg_pos NS fn i"
    proof (cases "fn \<in> set fs")
      case False
      with ws[unfolded ws'] show ?thesis by (auto simp: f)
    next
      case True
      let ?l = "Fun f (map var_x_i [0..<n])"
      let ?r = "var_x_i i"
      have ns: "(?l, ?r) \<in> NS" using orient i True unfolding ns_def f by force
      show ?thesis unfolding f
      proof (rule simple_arg_posI)
        fix ts :: "('a,string)term list"
        assume len: "length ts = n" and i: "i < n"
        define inv where "inv = (\<lambda>s. ts ! the_inv show (tl s))"
        {
          fix i
          have "inv (the_Var (var_x_i i)) = ts ! i" unfolding inv_def
            using the_inv_f_f[OF show_nat_inj] by auto
        }
        with subst.closedD[OF subst_NS ns, of inv]
        show "(Fun f ts, ts ! i) \<in> NS"
          by (auto simp: o_def len[symmetric] map_nth)
      qed
    qed
    show "simple_arg_pos NS fn i" by fact
  next
    assume ssimple: ssimple
    then obtain gs where sst': "?ss' = Some gs" unfolding ssimple_def by (cases ?ss', auto)
    {
      fix i fn 
      assume i: "i \<in> set (status \<sigma> fn)"
      obtain f n where f: "fn = (f,n)" by force
      show "simple_arg_pos S fn i"
      proof (cases "fn \<in> set gs")
        case False
        with sst[unfolded sst'] show ?thesis by (auto simp: f)
      next
        case True
        let ?l = "Fun f (map var_x_i [0..<n])"
        let ?r = "var_x_i i"
        have s: "(?l, ?r) \<in> S" using orient i True unfolding s_def f ssimple_comps_def sst' using ssimple by force
        show ?thesis unfolding f
        proof (rule simple_arg_posI)
          fix ts :: "('a,string)term list"
          assume len: "length ts = n" and i: "i < n"
          define inv where "inv = (\<lambda>s. ts ! the_inv show (tl s))"
          {
            fix i
            have "inv (the_Var (var_x_i i)) = ts ! i" unfolding inv_def
              using the_inv_f_f[OF show_nat_inj] by auto
          }
          with subst.closedD[OF subst_S s, of inv]
          show "(Fun f ts, ts ! i) \<in> S"
            by (auto simp: o_def len[symmetric] map_nth)
        qed
      qed
    }
    {
      obtain f :: 'a where True by auto
      define n where "n = 1 + max_list (map snd gs)" 
      have "(f,n) \<notin> set gs"
      proof
        assume "(f,n) \<in> set gs" 
        hence "n \<in> set (map snd gs)" by auto
        from max_list[OF this] show False unfolding n_def by simp
      qed
      from sst[unfolded sst', simplified, rule_format, OF this]
      have "simple_arg_pos S (f, n) 0" unfolding n_def by auto
      from this[unfolded simple_arg_pos_def, simplified, rule_format, of "replicate n undefined"]
      obtain a b where ab: "(a,b) \<in> S" unfolding n_def by auto
      thus "S \<noteq> {}" by auto
    }
    from ssimple obtain m where large_opt: "?large_opt = Some m" unfolding ssimple_def by (cases ?large_opt, auto)
    fix f g
    assume lf: "large f" 
    have large: "\<And> g. large g = (?pr g = m)" unfolding large_def large_opt using ssimple by auto
    note large_opt = large_opt[unfolded large_of_def Let_def]
    from large_opt have m: "m > 0" by (auto split: if_splits)
    have snd: "snd (?prc g f) = (?pr g \<ge> ?pr f)" for g f unfolding prc_nat_def Let_def by simp
    have fst: "fst (?prc g f) = (?pr g > ?pr f)" for g f unfolding prc_nat_def Let_def by simp
    from large_opt have max: "\<And> f. f \<in> set (map fst param) \<Longrightarrow> ?pr f \<le> m"
      unfolding ceta_map_of fun_of_map_fun'.simps
      by (force split: if_splits option.splits intro!: max_list) 
    have 0: "\<And> f. f \<notin> set (map fst param) \<Longrightarrow> ?pr f = 0" unfolding ceta_map_of fun_of_map_fun'.simps
      using map_of_eq_None_iff by force
    from large_opt have status: "\<And> f. f \<in> set (map fst param) \<Longrightarrow> ?pr f = m \<Longrightarrow> status \<sigma> f = []"
      by (auto split: if_splits)
    show "snd (?prc g f) \<Longrightarrow> large g" using max[of g] 0 lf unfolding large fst snd by fastforce
    thus "fst (?prc f g) \<or> snd (?prc f g) \<and> status \<sigma> g = []" using status[of g] max[of g] m 0
      unfolding large fst snd by (metis large lf status nat_le_linear nat_less_le)
  qed
  from af_compat have "af_compatible ?pi' NS" .
  from af_wpo_compat[OF this, of \<sigma>] have pi: "af_compatible ?pi NS"
    unfolding param d split by auto
  from af_wpo_status[of ?pi' \<sigma>] have pi2: "af_compatible_status ?pi \<sigma>"
    unfolding param d split by auto
  let ?m = "max_list (map (snd o fst) ?stat)"
  have \<sigma>: "\<exists>n. \<forall>m\<ge>n. \<forall>f. {0, 1} \<subseteq> set (status \<sigma> (f, Suc (Suc m)))"
  proof (rule exI[of _ ?m], intro allI impI)
    fix k f
    assume k: "?m \<le> k"
    let ?k = "Suc (Suc k)"
    from k have "?m < ?k" by auto
    with max_list[of ?k "map (snd o fst) ?stat"] 
    have "(f,?k) \<notin> fst ` set ?stat" by force
    from status_of_default[OF stat this]
    show "{0,1} \<subseteq> set (status \<sigma> (f,Suc (Suc k)))" by auto
  qed
  from wpo_ce_af_redtriple_order[OF \<sigma> pi pi2]
  have mono: "mono_ce_af_redtriple_order ?wpo_s ?wpo_ns ?wpo_ns ?pi" by simp
  interpret mono_ce_af_redtriple_order ?wpo_s ?wpo_ns ?wpo_ns ?pi by fact
  have mpi: "?mpi = empty_af" unfolding wpo_redtriple_def Let_def stat
    by simp
  interpret cpx_ce_af_redtriple_order ?wpo_s ?wpo_ns ?wpo_ns ?pi ?mpi ?Cpx unfolding mpi
    by (unfold_locales, auto simp: param d empty_af no_complexity_check_def)  
  have cpx: "cpx_ce_af_redtriple_order ?wpo_s ?wpo_ns ?wpo_ns ?pi ?mpi ?Cpx" ..
  let ?ws = "redtriple.not_wst ?rp"
  let ?ss = "redtriple.not_sst ?rp"
  have 1: "set s_list \<subseteq> ?wpo_s" using stri cb_all unfolding all_def by auto
  have 2: "set ns_list \<subseteq> ?wpo_ns" using nstri cb_all unfolding all_def by auto
  have 3: "set nst_list \<subseteq> ?wpo_ns" using nstrit cb_all unfolding all_def by auto
  have "not_subterm_rel_info ?wpo_ns (Some (map fst ?stat)) \<and> not_subterm_rel_info ?wpo_s (Some (map fst ?stat))" 
    unfolding not_subterm_rel_info.simps
  proof (intro allI impI conjI)
    fix fk i
    assume nmem: "fk \<notin> set (map fst ?stat)"
    obtain f k where f: "fk = (f,k)" by force
    with nmem have "(f,k) \<notin> fst ` set ?stat" by force
    from status_of_default[OF stat this]
      subterm_wpo_s_arg[of _ f] subterm_wpo_ns_arg[of _ f] 
    show "simple_arg_pos ?wpo_s fk i" "simple_arg_pos ?wpo_ns fk i" unfolding f
      by (intro simple_arg_posI, auto)+
  qed
  then have 4: "not_subterm_rel_info ?wpo_ns ?ws" 
    and 5: "not_subterm_rel_info ?wpo_s ?ss" by (auto simp add: param d)
  show "\<exists>S NS NST. cpx_ce_af_redtriple_order S NS NST ?pi ?mpi ?Cpx \<and>
                 set s_list \<subseteq> S \<and>
                 set ns_list \<subseteq> NS \<and>
                 set nst_list \<subseteq> NST \<and>
                 not_subterm_rel_info NS ?ws \<and>
                 not_subterm_rel_info S ?ss \<and>
                 (isOK (redtriple.mono ?rp (s_list @ ns_list @ nst_list)) \<longrightarrow>
                  mono_ce_af_redtriple_order S NS NST ?pi \<and> ctxt.closed S)"
  proof (intro exI conjI, rule cpx, rule 1, rule 2, rule 3, rule 4, rule 5, 
      intro impI conjI mono ctxt_closed_WPO_S)
    fix f m
    assume mono: "isOK (redtriple.mono ?rp (s_list @ ns_list @ nst_list))"
    show "set (status \<sigma> (f, m)) = {0..<m}" 
    proof (cases "map_of ?stat (f,m)")
      case None
      from status_of_default[OF stat this[unfolded map_of_eq_None_iff]] 
      show ?thesis by auto
    next
      case (Some idx)
      from map_of_SomeD[OF this] mono show ?thesis
        by (auto simp: d param status_of_Some[OF stat] Some)
    qed
  qed
qed

end