(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009, 2010)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Mktree
imports "../TermFun"
begin

section\<open>Definition of Huffman's algorithm\<close>

text \<open>taken from Nipkow's Isabelle presentation at TPHOLs'08\<close>

datatype 'a tree = Leaf nat 'a | Node nat "('a tree)" "('a tree)"
datatype_compat tree

primrec "value" where
"value (Leaf n _) = n" |
"value (Node n _ _) = n"

definition merge where
  "merge t1 t2 = Node (value t1 + value t2) t1 t2"

fun ins where
"ins u (t#ts) = (if value u \<le> value t then u # t # ts else t # ins u ts)" |
"ins u [] = [u]"

function (sequential) mkTree where
"mkTree(t1#t2#ts) = mkTree(ins (merge t1 t2) ts)" |
"mkTree[t] = t"
by pat_completeness auto




setup \<open>Term_Fun.register_datatypes [@{type_name tree}]\<close>
define_trs value.simps
define_trs merge_def
define_trs ins.simps

 (* here we can get away with underspecification,
since we do not need the simulation lemma *)
define_trs (no_sim) mkTree.psimps

thm mkTree.trs_def
termination mkTree
apply (trs_reduce mkTree)
apply expand_trs_problem
export_termination_problem "mkTree.xml"
apply (ceta_import_proof "proofs/mkTree.proof.xml")
apply eval
done

end
